indexing

	description: "ISE conversion of strings to pointers and vice versa."
	thanks: "The mico/E team for the idea."

	date: "$Date: 2003/11/29 $"
	revision: "$Revision: #8 $"

class

	EPX_STRING_HELPER


inherit

	ABSTRACT_STRING_HELPER
		redefine
			wipe_out,
			unfreeze_all
		end

	EPX_ISE_GC_HELPER


feature -- ISE specific conversions

	pointer_to_string (p: POINTER): STRING is
		do
			if p = default_pointer then
				Result := ""
			else
				create Result.make_from_c (p)
			end
		end

	set_string_from_pointer (s: STRING; p: POINTER) is
		do
			if p = default_pointer then
				s.clear_all
			else
				s.make_from_c (p)
			end
		end

	unfreeze_all is
		do
			precursor
			unfreeze_objects
		end


feature -- String utilities that have different names per compiler...

	wipe_out (s: STRING) is
			-- Remove all characters in `s', but do not change its
			-- capacity. Notably ISE doesn't implement ELKS correctly
			-- here.
		do
			s.clear_all
		end


feature {NONE} -- Implementation

	clear_unsaved_strings is
			-- The strings in `saved_strings' must be released to the
			-- garbage collector upon `unfreeze_all'.
		do
		end

	do_string_to_pointer (s: STRING): POINTER is
			-- Return a pointer to a linear area containing all the data
			-- in `s'. The area is zero-terminated.
			-- Pointers returned by this routine should only be used
			-- inside a single routine. They should not be kept alive
			-- longer than the duration of that single routine, and that
			-- single routine may not call any other routine which calls
			-- this function. Inside eposix this is guaranteed, but you
			-- are on your own if you decide to call this routine
			-- yourself.
			-- For moving garbage collectors you should regularly call
			-- `unfreeze_all', else memory will not be freed.
		local
			a: ANY
			buf: STDC_BUFFER
		do
			a := s.to_c
			if a /= Void then
				create buf.allocate (s.count + 1)
				buf.memory_copy ($a, 0, 0, s.count)
				-- Set zero byte
				buf.poke_character (buf.capacity - 1, '%U')
				frozen_objects.put_last (buf)
				Result := buf.ptr
			end
		ensure then
			one_more_frozen_object: frozen_objects.count = old frozen_objects.count + 1
		end

	do_uc_string_to_pointer (s: UC_STRING): POINTER is
			-- Return a pointer to a linear area containing all the data
			-- in the Unicode string `s'.
			-- Currently getting such a pointer is not an operation
			-- directly supported by UC_STRING and we need to do
			-- different things for certain compilers.
		local
			a: ANY
			buf: STDC_BUFFER
			real_s: STRING
		do
			real_s := s
			a := real_s.to_c
			if a /= Void then
				create buf.allocate (s.byte_count + 1)
				buf.memory_copy ($a, 0, 0, s.byte_count)
				-- Set zero byte
				buf.poke_character (buf.capacity - 1, '%U')
				frozen_objects.put_last (buf)
				Result := buf.ptr
			end
		ensure then
			one_more_frozen_object: frozen_objects.count = old frozen_objects.count + 1
		end


end
